/* Encyklopedie jazyka C
 *
 * Kapitola Preprocesor jazyka C
 *
 * 2023-08-12-19.53
 */

/* Copyright (c) David Martinek, 2023
 * david.martinek.78 (zavináč) gmail.com
 *
 * This work is licensed under a Creative Commons
 * Attribution-NonCommercial-ShareAlike 4.0 International License.
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 * Toto dílo je licencováno pole licence Creative Commons
 * Uveďte původ-Neužívejte dílo komerčně-Zachovejte licenci 4.0 Mezinárodní
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/deed.cs
 */

#define VERZE "1.8"
#define NAZEV "Preprocesor jazyka C"

// TODO Prostuduj si obsah tohoto hlavičkového souboru! Jsou tam zaklínadla,
// bez kterých některé věci ve Windows nefungují správně!
// - barevný výstup
// - printg - český tisk (překódovává z UTF8 na CP852)
// - printh - český tisk + vyprávěcí mód
// - formátovací a formátové značky pro printf/scanf podle normy
// - obsluha nabídek v kapitolách
#include "pomocny-modul.h"

// Implementace nabídkového menu
#include "menu.h"

#include <stdio.h>
#include <stdlib.h>
// #include <string.h>
#include <stdbool.h>
// #include <ctype.h>
// #include <math.h>
// #include <float.h>
// #include <limits.h>


// Hlavičky funkcí realizující jednotlivé kapitoly.

void coJeToPreprocesor(void);
void komentare(void);
void vkladaniSouboru(void);
void podminenyPreklad(void);
void makraBezParametru(void);

void makraSParametry(void);
void diagnostika(void);
void parametryMaker(void);
void direktivyPrekladace(void);
//void test(void);


int main(void)
{
  povolBarevnyVystup();
//  prepniRychlostVypraveni(2);

  Tmenu * menu = menuVytvor(
    NADPIS(
      "==============================================================\n"
      "Encyklopedie jazyka C - " NAZEV " (verze " VERZE ")\n"
      "(c) David Martinek, 2023\n"
      "Zároveň s programem sleduj i zdrojový text!\n"
      "==============================================================\n"
    )
    , zobrazMod
  );

  if (menu == NULL)
  {
    printg("Není dost paměti ani pro výrobu menu. Končím program.");
    return 0;
  }

  menuPridej(menu,   0, "\n" SEKCE("Základní použití") , NULL);
  menuPridej(menu, 'A', "Co je to preprocesor\n", coJeToPreprocesor);
  menuPridej(menu, 'B', "Komentáře\n", komentare);
  menuPridej(menu, 'C', "Vkládání souborů\n", vkladaniSouboru);
  menuPridej(menu, 'D', "Podmíněný překlad\n", podminenyPreklad);
  menuPridej(menu, 'E', "Makra bez parametrů - symbolické konstanty\n", makraBezParametru);

  menuPridej(menu,   0, "\n" SEKCE("Expertní použití") , NULL);
  menuPridej(menu, 'F', "Makra s parametry\n", makraSParametry);
  menuPridej(menu, 'G', "Kouzla s parametry\n", parametryMaker);
  menuPridej(menu, 'H', "Diagnostika\n", diagnostika);
  menuPridej(menu, 'I', "Pragma direktivy překladače\n", direktivyPrekladace);
//  menuPridej(menu, 'J', "\n", test);
//  menuPridej(menu,   0, "\n" SEKCE("Druhá sekce") , NULL);
//  menuPridej(menu, 'K', "\n", test);
//  menuPridej(menu, 'L', "\n", test);

  menuPridej(menu,   0, "\n" SEKCE("Nastavení programu"), NULL);
  menuPridej(menu, '0', "Nastav vyprávěcí mód\n", nastavRychlostVypraveni);

  while (menuVolba(menu) != 'Q') {}

  printg("Končím program.\n");
  menuZrus(menu);
  return 0;
}




////////////////////////////////////////////////////////////////////////////////
void coJeToPreprocesor(void)
{
  zacatekKapitoly("Co je to preprocesor");

  printh(
    "- Je to program pro automatizované provádění textových náhrad v kódu.\n"
    "- Spouští se ještě před kompilátorem.\n"
    "\n"
    "- V balíku programů GCC (na Windows MinGW) je preprocesor program\n"
    "  s názvem " BB("cpp") " (C PreProcessor).\n"
    "\n"
  );

  if (navratNeboSekce("Oblasti použití preprocesoru")) return;
  if (pokracovat())
  {
    printh(
      PODSEKCE("Běžné použití")
      "- Komentáře\n"
      "- Symbolické konstanty (příkaz " KK("#define") ")\n"
      "- Vkládání hlavičkových souborů (příkaz " KK("#include") ")\n"
      "- Ošetření hlavičkových souborů (příkaz " KK("#ifndef") ")\n"
      "\n"
      PODSEKCE("Pokročilé použití")
      "- Podmíněný překlad\n"
      "- Makra\n"
      "- Diagnostika v době překladu\n"
      "- Direktivy překladače (příkaz " KK("#pragma") ")\n"
      "\n"
    );
  }

  if (navratNeboSekce("Lexikální preprocesor")) return;
  if (pokracovat())
  {
    printh(
      "- Preprocesor jazyka C je tzv. lexikální preprocesor.\n"
      "- To znamená, že preprocesor vůbec neřeší syntaxi jazyka C.\n"
      "- Dělá textové náhrady pouze celých jednotlivých lexémů - tedy slov.\n"
      "\n"
      KOD(
      "#define AB 1.2\n"
      "const int ABCD = AB+3.4;\n"
      "char * text = \"Hodnota ABCD je AB + 3.4\";\n"
      )
      "\n"
      "- Za hranice slov považuje bílé znaky, operátory jazyka C, závorky apod.\n"
      "- Proto neprovede náhradu ve slově ABCD, ale ve výrazu ano.\n"
      "- Nedělá náhrady uprostřed stringů. \n"
      "- Preprocesor zpracuje předchozí kód takto:\n"
      "\n"
      KOD(
      "const int ABCD = 1.2+3.4;\n"
      "char * text = \"Hodnota ABCD je AB + 3.4\";\n"
      )
      "\n"
    );
  }

  if (navratNeboSekce("Přenositelnost")) return;
  if (pokracovat())
  {
    printh(
      "- Byl zamýšlen jako nástroj pro zjednodušení programátorské práce.\n"
      "- Jeho použitím mohl zůstat jazyk C i jeho překladač velmi malý a efektivní.\n"
      "- Zároveň umožňuje relativně snadno přizpůsobovat zdrojový kód různým\n"
      "  cílovým platformám a pomáhá jej tak dělat přenositelným.\n"
      "- Pomocí podmíněného překladu jde testovat, na jaké platformě je program\n"
      "  překládán a přizpůsobit tomu kód.\n"
      "- To je ostatně vidět i na zdrojovém kódu modulů tohoto projektu.\n"
      // Například funkce printg a printh fungují na Linuxu a ve Windows vnitřně
      // jinak, ale díky preprocesoru to v modulu main.c není nijak poznat.
      "\n"
//      "- Z historických důvodů používá preprocesor i jazyk C++.\n"
//      "- Nově vznikající jazyky preprocesor nepoužívají, protože se jeho použití\n"
//      "  dnes jeví jako potenciálně nebezpečné.\n"
//      // Například nevhodně vytvořená makra vytváří nepříjemné chyby.
//      // Autoři dnešních jazyků také příliš neřeší HW nároky dnešních překladačů.
//      "- Novější jazyky dosahují přenositelnosti na jiné platformy vysokoúrovňovým\n"
//      "  datovým modelem za cenu horší efektivity.\n"
//      // To znamená mnohem větší nároky na tvorbu překladače a horší efektivitu
//      // výsledného kódu. To je také důvod, proč je překladač jazyků C/C++
//      // na nových HW platformách většinou dostupný jako první.
//      "\n"
    );
  }

  if (navratNeboSekce("Dobrý sluha, zlý pán")) return;
  if (pokracovat())
  {
    printh(
      "- Preprocesor lze vnímat jako nástroj pro řízení překladu.\n"
      "- Jako nástroj pro algoritmizaci je nebezpečný.\n"
      "- Pokud jde problém řešit jinak, je dobré se preprocesoru vyhýbat.\n"
      "\n"
      "- Vyhýbejte se makrům s parametry.\n"
      "  => Při nevhodném použití vytváří zákeřné chyby.\n"
      "  => Vyšší efektivita oproti funkcím nestojí za to - překladače dnes\n"
      "     umí vytvářet velice optimalizovaný kód samy.\n"
      "  => Místo maker jde používat i " KK("static inline") " funkce.\n"
      "\n"
      "- Typová kontrola se dělá až po použití maker, ne v místě jejich tvorby.\n"
      "  => Chyby při překladu se mohou objevit na nečekaných místech.\n"
      "\n"
      "- Makra mohou způsobit znečištění jmenného prostoru.\n"
      "  => Textové náhrady mohou nechtěně vyrobit již použité identifikátory.\n"
      "  => To vede ke kolizím identifikátorů.\n"
      "  => Ještě horší je nevědomé překrytí identifikátorů. To nezpůsobí chybu\n"
      "     při překladu, ale nezamýšlené chování.\n"
      "\n"
      /* V tomto projektu jsou makra používána v podstatě jako značkovací jazyk
      pro značení nadpisů, kódu a dalších elementů, které chci mít na výstupu
      barevně zvýrazněné. Kdybych tato makra udělal jako funkce, nešlo by je
      používat přímo jako součást textových řetězců.

      Využívám vlastnosti textových řetězců v C, kdy "toto" "a toto" tvoří jeden
      stringový literál. Protože moje makra vrací textový literál, můžu takové
      makro použít mezi dvěma stringy a dojde k jejich spojení. To by pomocí
      podprogramů nešlo.
      */
    );
  }

  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void komentare(void)
{
  zacatekKapitoly("Komentáře");

  printh(
    "- Preprocesor odstraňuje komentáře ze zdrojového textu.\n"
    "- Každý komentář je nahrazen jednou mezerou.\n"
    "- Překladač tak dostane na vstup výrazně kratší text obsahující pouze kód.\n"
    "\n"
    "- Existují řádkové (" KK("//") ") a blokové komentáře (" KK("/* */") ").\n"
    "- Blokové mohou zakomentovat celý blok, tedy i více řádků kódu.\n"
    "- Hodí se i na zakomentování malé části uvnitř jediného řádku.\n"
    "- S rozvojem kvalitních editorů se i celé bloky kódu dnes komentují\n"
    "  klávesovými zkratkami pomocí řádkových komentářů. Je to vizuálně\n"
    "  výraznější.\n"
    // V Code::Blocks Ctrl+Shift+C a Ctrl+Shift+X, jinde často Ctrl+/.
    "\n"
    KOD(
    "  // Řádkový komentář končí na konci řádku\n"
    "  /* Blokový komentář může mít\n"
    "    i více řádků.\n"
    "  */\n"
    )
    "\n"
  );

  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void vkladaniSouboru(void)
{
  zacatekKapitoly("Vkládání souborů");

  printh(
    "- V místě použití příkazu " KK("#include") " se vloží obsah zadaného souboru.\n"
    "- Typicky se vkládají hlavičkové soubory.\n"
    "- Jde vkládat i jiné textové soubory, ale nedoporučuje se to.\n"
    "\n"
    KOD(
    "  #include <stdio.h>\n"
    "  #include \"pomocny-modul.h\"\n"
    )
    "\n"
    "- Varianta s úhlovými závorkami hledá soubor v systémovém adresáři.\n"
    "- Varianta s uvozovkami hledá soubor v aktuální adresáři projektu.\n"
    "- Blíže je toto probíráno v kapitole Moduly.\n"
    "\n"
  );

  if (navratNeboSekce("Ošetření hlavičkového souboru")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "#ifndef __SOUBOR_H__\n"
      "#define __SOUBOR_H__\n"
      "// ... obsah hlavičkového souboru\n"
      "#endif\n"
      )
      "\n"
      "- Hlavičkové soubory se ošetřují pomocí podmíněného překladu (viz dále).\n"
      "- Takto musí být ošetřen každý hlavičkový soubor.\n"
      "- Zabraňuje to opakovanému vložení hlavičkového souboru - podrobněji\n"
      "  viz kapitolu Moduly.\n"
      "- Při prvním průchodu dojde k definování symbolu " KK("__SOUBOR_H__") ".\n"
      "- Při opakovaném vkládání stejného souboru už je symbol definován\n"
      "  a podmínka zabrání opětovnému vložení.\n"
      "- Název symbolu musí být volen tak, aby pokud možno nekolidoval s nějakým\n"
      "  jiným symbolem, který uživatel může chtít používat.\n"
      // Například symboly MAX, NAME apod. by určitě někde kolidovaly
      // s uživatelskými konstantami.
      // Doporučuje se používat podtržítka, případně i náhodné skupiny znaků:
      // __MUJMODUL_H_MLKSDVLMASDFRP4ULKRJS09__ je mnohem bezpečnější název.
      // Na druhou stranu, není potřeba to přehánět.
      "\n"
    );
  }

  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void podminenyPreklad(void)
{

  zacatekKapitoly("Podmíněný překlad");

  printh(
    "- Umožňuje měnit kód na základě testování různých podmínek.\n"
    "- Například jde vytvářet kód pro různé cílové platformy.\n"
    "- Nejčastějším použitím je ošetření hlavičkových souborů proti vícenásobnému\n"
    "  vložení.\n"
    "\n"
  );

  if (navratNeboSekce("Příkazy pro podmíněný překlad")) return;
  if (pokracovat())
  {
    printh(
      "- Preprocesor má několik podmíněných příkazů, jimiž se řídí překlad.\n"
      "- Funguje to tak, že označený kus kódu se zahrne nebo vymaže na základě\n"
      "  vyhodnocení podmínky ještě před samotným překladem.\n"
      "\n"
      "  " KK("#if podmínka") "\n"
      "  - Zahrne následující text, když je podmínka pravdivá.\n"
      "  " KK("#elif podmínka") "\n"
      "  - Další větev s vyhodnocením podmínky.\n"
      "  " KK("#ifdef symbol") "\n"
      "  - Zahrne následující text, když je zadaný symbol definován.\n"
      "  " KK("#ifndef symbol") "\n"
      "  - Zahrne následující text, když zadaný symbol definován není.\n"
      "  " KK("#else") "\n"
      "  - Větev else - musí následovat po některém z předchozích příkazů.\n"
      "  " KK("#endif") "\n"
      "  - Ukončuje podmíněný překlad. Všechny předchozí příkazy jím musí končit.\n"
      "\n"
      "- " KK("Symbol") " je název makra vyrobený pomocí " KK("#define") ", nebo\n"
      "  předaný preprocesoru při spuštění přepínačem " KK("-Dsymbol") " z příkazové řádky.\n"
      "\n"
      "- " KK("Podmínka") " je aritmetická podmínka se stejnými pravidly jako v jazyce C.\n"
      "- Jde v ní použít běžné aritmetické (+, -, *, /), bitové (^, <<, >>, ~),\n"
      "  logické (||, &&) a porovnávací (<, >, <=, >=, ==, !=) operátory.\n"
      "- Konstrukcí " KK("defined symbol") " jde testovat existenci symbolu.\n"
      "- Jde testovat existenci a hodnoty symbolů (maker) preprocesoru, ne jazyka C.\n"
      "- Neumí to testovat hodnoty proměnných, " KK("sizeof") ", ani nic podobného.\n"
      "- Hodnoty použitých symbolů (maker) musí být čísla nebo znaky.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka - multiplatformní kód")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "  int e = -1;\n"
      "#ifdef __unix__\n"
      "  e = system(\"clear\");\n"
      "#elif defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64\n"
      "  e = system(\"cls\");\n"
      "#endif\n"
      "  if (e != 0)\n"
      "    printf(\"\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\\n\");\n"
      )
      "\n"
      "- V této ukázce se testuje existence symbolů (maker) definovaných\n"
      "  za účelem detekce operačního systému.\n"
      "- Cílem této ukázky je vyčistit konzoli, což se ale v různých operačních\n"
      "  systémech dělá různými způsoby.\n"
      // Seznam předdefinovaných symbolů překladače jde vypsat pomocí
      // echo | gcc -dM -E -
      "\n"
    );
  }

  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void makraBezParametru(void)
{
  zacatekKapitoly("Makra bez parametrů - symbolické konstanty");

  printh(
    "- Používají se pro symbolické konstanty.\n"
    "- GNU manuál o nich hovoří jako o \"object-like macros\".\n"
    "- Tvoří se příkazem " KK("#define") ".\n"
    "\n"
    KOD(
    "  #define PI 3.141592653589793\n"
    )
    "\n"
    "- Nyní bude ve zdrojovém textu každý výskyt symbolu PI, nahrazen textem\n"
    "  3.141592653589793.\n"
    "\n"
  );

  if (navratNeboSekce("Proč nepoužívat středník")) return;
  if (pokracovat())
  {
//    #include <math.h>
//    #define PI 3.141592653589793;
//    float y = sin(PI);

    printh(
      "- Začínající programátory v C to svádí používat za definicí symbolu středník.\n"
      "- Je třeba pamatovat, že symbolické konstanty má na starosti preprocesor.\n"
      "- Případný středník dělá potíže při použití symbolu ve výrazech.\n"
      "\n"
      KOD(
      "// středník je chyba!\n"
      "#define PI 3.141592653589793;\n"
      "float y = sin(PI);\n"
      )
      "\n"
      "- Po zpracování preprocesorem dostane překladač tento kód:\n"
      "\n"
      KOD(
      "float y = sin(3.141592653589793;);\n"
      )
      "\n"
      "- Přebytečný středník tady způsobí syntaktickou chybu.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Pojmenované výrazy")) return;
  if (pokracovat())
  {
    printh(
      "- Tělem makra může být jakýkoli text.\n"
      "- Obecně je takové použití velmi nebezpečné a matoucí.\n"
      "\n"
      KOD(
      "#define A printf\n"
      "#define B (\"Baf!\\n\")\n"
      "A B;\n"
      )
      "=> printf (\"Baf!\\n\");\n"
      "=> "
    );
    #define A printf
    #define B ("Baf!\n")
    A B;


    #define VYRAZA 1 + 2
    #define VYRAZB (1 + 2)

    printh(
      "\n"
      "- Kromě literálů dávají smysl jen výrazy s konstantním výsledkem.\n"
      "- Pro bezpečnou použitelnost platí jedno pravidlo. => Závorkovat!\n"
      "- Důležitost se projeví při použití ve výrazech.\n"
      "\n"
      KOD(
      "#define VYRAZA 1 + 2\n"
      "#define VYRAZB (1 + 2)\n"
      )
      "\n"
      "=> VYRAZA * 2 => 1 + 2 * 2 = %d\n"
      "=> VYRAZB * 2 => (1 + 2) * 2 = %d\n"
      "\n"
      , VYRAZA * 2
      , VYRAZB * 2
    );
  }

  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void makraSParametry(void)
{
/*
Makra
- Je lepší se jim vyhnout.
- Při volání vypadají jako funkce, ale nejde o funkce.
- Neprovádějí typovou kontrolu.
- Je potřeba si dávat pozor, co se předá jako argument.

- Makra bez parametrů
- Makra s parametry
- Na co si dávat pozor

- Alternativy
  - Nové verze jazyka C, jako ISO C99, C11, C23 poskytují dostatek nástrojů,
    které potřebu nebezpečných použití preprocesoru eliminují.
  - Makra jde v modulech nahradit pomocí static inline funkcí.
  - Symbolické konstanty jde nahradit pomocí konstantních proměnných.
    - Nejde to ale všude - například u deklarace pole je rozdíl, když se pro
      specifikaci rozměru dimenze použije symbolická konstanta a konstantní
      proměnná.
*/
  zacatekKapitoly("Makra s parametry");

  printh(
    "- Vypadají jako funkce, ale pořád jsou to textové náhrady.\n"
    "- Výsledkem aplikace makra je kód.\n"
    "- Parametry makra jsou výrazy, které se dosazují do generovaného kódu.\n"
    "- Nevhodným používáním může snadno vzniknout chybný kód nebo syntaktická chyba.\n"
    "\n"
    "- Nelze vyrábět rekurzivní makra.\n"
    "- Makro generuje kód v době před překladem.\n"
    "- V době běhu programu se provádí kód, který je výsledkem provedení makra.\n"
    // Pokud si tento rozdíl neuvědomíte a používáte makra místo funkcí, mohou
    // vás některé věci zaskočit. Např. proč __TIME__ nevrací aktuální čas
    // za běhu programu.
    "\n"
    "- Nepoužívat, dokud nevíte co děláte!\n"
    "- Pokud si myslíte, že víte, co děláte, stejně je nepoužívejte!\n"
    // Pravděpodobně to bude jenom D-K efekt... ;-)
    "\n"
  );

  if (navratNeboSekce("Syntaxe")) return;
  if (pokracovat())
  {
/*
    #define max(a,b) (a) > (b) ? (a) : (b)

    #define max(a,b) \
    (a) > (b) ? \
      (a) : \
      (b)
*/

    printh(
      "- Ideální jsou makra, která se vejdou na jediný řádek.\n"
      "\n"
      KOD(
      "  #define max(a,b) (a) > (b) ? (a) : (b)\n"
      )
      "\n"
      "- Pro rozdělení kódu na více řádků jde použít znak opačného lomítka.\n"
      "- Tento symbol se nepoužívá na posledním řádku.\n"
      "\n"
      KOD(
      "#define max(a,b) \\\n"
      "(a) > (b) ? \\\n"
      "  (a) : \\\n"
      "  (b)\n"
      )
      "\n"
    );
  }

  if (navratNeboSekce("Bezpečnostní pravidla")) return;
  if (pokracovat())
  {
    printh(
      "- Makra se používají podobně jako podprogramy, takže by se podle toho\n"
      "  měla chovat.\n"
      "- Uživatel si nemusí být vědom, že používá makro místo funkce.\n"
      "\n"
      "1. Na konci makra se nepíše středník.\n"
      "  - Ten píše až uživatel při použití makra.\n"
      "2. Jako parametr musí jít dosadit výraz.\n"
      "3. Makro musí reprezentovat právě jeden výraz, nebo příkaz.\n"
      "\n"
      "- Nedodržení těchto pravidel vede k zákeřným chybám.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Parametrem může být výraz")) return;
  if (pokracovat())
  {
    #define krat1(a,b) a * b
    #define krat2(a,b) (a) * (b)

    printh(
      "- Řešením je důsledně závorkovat parametry v těle makra.\n"
      "- Je-li tělem makra výraz, i ten by měl být v závorkách.\n"
      "\n"
      KOD(
      "#define krat1(a,b) a * b\n"
      "#define krat2(a,b) ((a) * (b))\n"
      )
      "\n"
      "=> krat1(1+2, 3+4) => 1+2*3+4 => %d\n"
      "=> krat2(1+2, 3+4) => ((1+2)*(3+4)) => %d\n"
      "\n"
      , krat1(1+2, 3+4)
      , krat2(1+2, 3+4)
    );
  }

  if (navratNeboSekce("Jeden příkaz - pokus č. 1")) return;
  if (pokracovat())
  {
//    #define POKUS1(a, b) printf("%d ", (a)); printf("%d", (b))
//    int a = 1, b = 2;
//    if (a < b)
//      POKUS1(a, b);
//    else // syntaktická chyba
//      POKUS1(b, a);

    printh(
      "- Při použití musí jít za makro napsat středník, aniž by to dělalo problémy.\n"
      "- Uživatel může chtít makro použít například jako větev příkazu " KK("if-else") ",\n"
      "  což musí být právě jeden příkaz.\n"
      "\n"
      KOD(
      "  #define POKUS1(a, b) printf(\"%%d \", (a)); printf(\"%%d\", (b))\n"
      "  int a = 1, b = 2;\n"
      "  if (a < b)\n"
      "    POKUS1(a, b); // => printf(\"%%d \", (a)); printf(\"%%d\", (b));\n"
      "  else            // syntaktická chyba\n"
      "    POKUS1(b, a); // => printf(\"%%d \", (b)); printf(\"%%d\", (a));\n"
      )
      "\n"
      "- Toto řešení způsobí syntaktickou chybu, protože druhý příkaz už\n"
      "  do větve příkazu " KK("if") " nepatří.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Jeden příkaz - pokus č. 2")) return;
  if (pokracovat())
  {
//    #define POKUS2(a, b) { printf("%d ", (a)); printf("%d", (b)); }
//    int a = 1, b = 2;
//    if (a < b)
//      POKUS2(a, b);
//    else // syntaktická chyba
//      POKUS2(b, a);

    printh(
      "- S řešením problému nepomohou ani složené závorky.\n"
      "- Středník za složenou závorkou se bere jako samostatný příkaz.\n"
      "\n"
      KOD(
      "  #define POKUS2(a, b) { printf(\"%%d \", (a)); printf(\"%%d\", (b)); }\n"
      "  int a = 1, b = 2;\n"
      "  if (a < b)\n"
      "    POKUS2(a, b); // => { printf(\"%%d \", (a)); printf(\"%%d\", (b)); };\n"
      "  else            // syntaktická chyba\n"
      "    POKUS2(b, a); // => { printf(\"%%d \", (b)); printf(\"%%d\", (a)); };\n"
      )
      "\n"
      "- Toto řešení způsobí opět syntaktickou chybu, protože středník je\n"
      "  druhý příkaz za blokovým příkazem a do větve příkazu " KK("if") " nepatří.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Řešení - do-while(0)")) return;
  if (pokracovat())
  {
    printh(
      "- Řešením je použít místo blokového příkazu konstrukci " KK("do {} while(0)") ".\n"
      "- Uvnitř může být posloupnost příkazů.\n"
      "- Když se zakončí středníkem, bude to právě jeden příkaz.\n"
      "\n"
      KOD(
      "  #define POKUS3(a, b) do { printf(\"%%d \", (a)); printf(\"%%d\", (b)) } while(0)\n"
      "  int a = 1, b = 2;\n"
      "  if (a < b)\n"
      "    POKUS3(a, b);\n"
      "  else            // Teď je to v pořádku.\n"
      "    POKUS3(b, a);\n"
      )
      "\n"
      "  => "
    );
    #define POKUS3(a, b) do { printf("%d ", (a)); printf("%d", (b)); } while(0)
    int a = 1, b = 2;
    if (a < b)
      POKUS3(a, b);
    else            // Teď je to v pořádku
      POKUS3(b, a);

    printf("\n\n");
  }


  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void parametryMaker(void)
{
  zacatekKapitoly("Kouzla s parametry maker");

  printh(
    "- Parametr makra je text a preprocesor s ním tak zachází.\n"
    "- Aplikací makra vzniká kód, takže parametr makra je text reprezentující\n"
    "  smysluplný výraz, který se do vytvářeného kódu dosadí.\n"
    "- V tomto oddíle se podíváme na pokročilá kouzla s parametry.\n"
    "\n"
  );

  if (navratNeboSekce("Stringifikace parametrů")) return;
  if (pokracovat())
  {
    #define TEXTPARAM(text) "Parametr text obsahuje " #text "."

    printh(
      "- Preprocesor si běžně nevšímá obsahu stringů.\n"
      "- Občas je ale potřeba parametr makra do stringu dosadit.\n"
      "- Případně je potřeba udělat z parametru string.\n"
      "- Hodí se to, když je parametrem nějaký výraz, který chci vytisknout jako\n"
      "  součást chybového hlášení - viz část o využití pro diagnostiku.\n"
      "\n"
      "- Z parametru se udělá string tak, že se před něj zapíše znak mřížky (#).\n"
      "- Pokud se má tento nový string stát součástí jiného stringu, je potřeba\n"
      "  jej zapsat vedle něj. Využívá to pravidlo, že dva stringové literály\n"
      "  oddělené bílými znaky se v jazyce C sloučí do jednoho.\n"
      "\n"
      KOD(
      "#define TEXTPARAM(text) \"Parametr text obsahuje \" #text \".\"\n"
      )
      "\n"
      "- Za parametr " KK("text") " jde při použití dosadit i víceslovný text.\n"
      "- Posloupnosti bílých znaků, včetně nových řádků, se ale vždy převedou\n"
      "  na právě jednu mezeru.\n"
      "\n"
      KOD(
      "TEXTPARAM(slovo)\n"
      )
      "=> \"Parametr text obsahuje \" \"slovo\" \".\" => %s\n"
      "\n"
      KOD(
      "TEXTPARAM(dvě            slova)\n"
      )
      "=> %s\n"
      "\n"
      KOD(
      "TEXTPARAM(x   >   (3  * a +           y))\n"
      )
      "=> %s\n"
      "\n"
      , TEXTPARAM(slovo)
      , TEXTPARAM(dvě            slova)
      , TEXTPARAM(x   >   (3  * a +           y))
    );
  }

  if (navratNeboSekce("Spojování tokenů")) return;
  if (pokracovat())
  {
    #define SPOJ(param1, param2) param1 ## param2

    printh(
      "- Někdy je potřeba spojit dva tokeny (typicky parametry) dohromady.\n"
      "- Výsledkem musí být smysluplný token (identifikátor, literál, operátor).\n"
      "- Je-li výsledkem něco nesmyslného, dojde k chybě.\n"
      "- Dělá se to zápisem dvou mřížek (##) mezi spojované tokeny.\n"
      "\n"
      KOD(
      "#define SPOJ(param1, param2) param1 ## param2\n"
      )
      "\n"
      "- Kdybychom pouze zapsali " KK("param1param2") " preprocesor sem nic nedosadí\n"
      "  a výsledkem by byl text " KK("param1param2") ".\n"
      "- Při zápisu " KK("param1 param2") " by ve výsledku byla mezera.\n"
      "- Operátor " KK("##") " spojí obsah obou parametrů do jednoho výsledku.\n"
      "\n"
      "- Příklady použití:\n"
      "  - výroba identifikátorů podle vzoru (předpona + parametr + přípona),\n"
      "  - výroba inicializátorů složitějších struktur,\n"
      "  - výroba generického kódu.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Ukázka spojování tokenů")) return;
  if (pokracovat())
  {
    #define ZMENA(co, jak, okolik) co jak ## = (okolik)

    int x = 5, y = 1;
    ZMENA(x, +, 10);
    ZMENA(y, <<, 4+4);

    printh(
      "\n"
      KOD(
      "#define ZMENA(co, jak, okolik) co jak ## = (okolik)\n"
      )
      "\n"
      "- V této ukázce se parametr " KK("jak") " spojí s " KK("=") " a vytvoří tak nový operátor.\n"
      "- Za parametr " KK("co") " se dosadí identifikátor proměnné.\n"
      "- Za parametr " KK("jak") " se dosadí aritmetický operátor\n"
      "- Za parametr " KK("okolik") " se dosadí hodnota, o jakou se má proměnná změnit.\n"
      "\n"
      KOD(
      "  int x = 5, y = 1;\n"
      )
      "  " KK("ZMENA(x, +, 10);") " => x += (10); => x: %d\n"
      "  " KK("ZMENA(y, <<, 4+4);") " => y <<= (4+4); => y: %d\n"
      "\n"
      , x
      , y
    );
  }


  konecKapitoly(__func__);
}




////////////////////////////////////////////////////////////////////////////////
void diagnostika(void)
{
  zacatekKapitoly("Diagnostika");

  printh(
    "- Makra se dají použít jako pomůcka při ladění a testování programů.\n"
    "- Mohou obsahovat testovací kód a generovat chybová hlášení.\n"
    "- Dokáží pracovat s údaji jako je číslo aktuálního řádku, název aktuální\n"
    "  funkce, či jméno aktuálního souboru.\n"
    "  => To je užitečné pro generování výstižných chybových hlášení.\n"
    "\n"
    "- V tomto oddíle si popíšeme konstrukce, které jsou pro tuto činnost užitečné.\n"
    "\n"
  );

  if (navratNeboSekce("Makro assert")) return;
  if (pokracovat())
  {
    printh(
      "- Toto makro slouží pro testování invariantních logických podmínek.\n"
      "- To jsou podmínky, o kterých si autor kódu myslí, že budou pravdivé\n"
      "  za jakýchkoli okolností.\n"
      "- Porušení takové podmínky je pak zjevně programátorská chyba.\n"
      "- Při porušení takové podmínky, dojde k zastavení programu a výpisu\n"
      "  chybového hlášení.\n"
      "- Používá se to při testování.\n"
      "- Pokud během testování makro nikdy nesepne, je naděje, že program funguje\n"
      "  správně.\n"
      "\n"
      "- Nachází se v assert.h.\n"
      "- Dá se vypnout příkazem " KK("#define NDEBUG 1") " před natažením hlavičkového\n"
      "  souboru assert.h.\n"
      "  => Makro se pak rozgeneruje na jednu mezeru.\n"
      "  => Po odladění není třeba testovací kód mazat nebo zakomentovat.\n"
      "\n"
      KOD(
      "  // pokud má file hodnotu NULL, je to programátorská chyba\n"
      "  assert(file != NULL);\n"
      )
      "\n"
    );

    // #define NDEBUG 1

    // Odkomentuj, přelož a spusť jednou se zakomentovaným předchozím řádkem
    // a podruhé jej odkomentuj.
    // #include <assert.h>
    // assert(1 > 2);
  }

  if (navratNeboSekce("Chybová hlášení a varování")) return;
  if (pokracovat())
  {
#warning "Příkaz #warning + -pedantic => další varování."
//#error "A tady končíme!"

#pragma message "TODO: Tohle je provizorní řešení. Oprav to!"
#pragma GCC warning "Pragma GCC warning + -pedantic => bez dalšího varování."
//#pragma GCC error "Tohle přeložit nejde!"

    printh(
      "\n"
      KOD(
      "  #warning \"Tento hlavičkový soubor je zastaralý!\"\n"
      "  #error \"Tento operační systém není podporován!\"\n"
      )
      "\n"
      "- Pokud se při překladu narazí na tyto konstrukce, dojde k výpisu\n"
      "  varování nebo chybového hlášení překladače.\n"
      "- Typicky se to používá spolu s podmíněným překladem.\n"
      "\n"
      "- Příkaz " KK("#warning") " je rozšířením GCC, které se snad stane standardem.\n"
      "- Stejně jde použít direktivy překladače GCC (nejsou přenositelné).\n"
      "- Takto vytvořené varování neprodukuje hlášení o nekompatibilitě s ISO C.\n"
      "- Příkaz " KK("#pragma message") " se hodí pro TODO poznámky, protože\n"
      "  se zobrazují jako zprávy při překladu.\n"
      "\n"
      KOD(
      "  #pragma message \"TODO: Tohle je provizorní řešení. Oprav to!\"\n"
      "  #pragma GCC warning \"Používání maker vede ke špatnému hodnocení v kurzech programování\"\n"
      "  #pragma GCC error \"Nechci, aby tento modul šel přeložit.\"\n"
      )
      "\n"
    );
  }

  if (navratNeboSekce("Makra pro hlášení chyb")) return;
  if (pokracovat())
  {
    #define HLASENI(text)\
      printf("%s\nSoubor: %s\nRadek: %d\nFunkce: %s\n", text, __FILE__, __LINE__, __func__)

    printh(
      KK("__FILE__") "\n"
      "- Řetězcový literál s názvem aktuálního souboru včetně cesty.\n"
      "\n"
      KK("__LINE__") "\n"
      "- Číslo řádku, na kterém bylo makro použito.\n"
      "\n"
      KK("__func__") "\n"
      "- String s názvem aktuální funkce.\n"
      "- Není to makro, ale proměnná zavedená od ISO C99.\n"
      "\n"
      KOD(
      "#define HLASENI(text)\\\n"
      "  printf(\"%%s\\nSoubor: %%s\\nRadek: %%d\\nFunkce: %%s\\n\", text, __FILE__, __LINE__, __func__)\n"
      "\n"
      "HLASENI(\"Pozor! Neco se deje!\");\n"
      )
      "\n"
      "=>\n"
    );

    HLASENI("Pozor! Neco se deje!");

    printf("\n");
  }

  if (navratNeboSekce("Makra s datem a časem")) return;
  if (pokracovat())
  {
    printh(
      "- Tato makra jsou užitečná pro zaznamenání data a času překladu.\n"
      "- Údaje se vztahují k době běhu preprocesoru.\n"
      "- Nedají se použít pro zjištění aktuálního data a času za běhu programu.\n"
      "- Makra vracejí stringový literál.\n"
      "\n"
      KK("__DATE__") " => " __DATE__ "\n"
      KK("__TIME__") " => " __TIME__ "\n"
      "\n"
    );
  }

  if (navratNeboSekce("Předdefinované symboly operačních systémů")) return;
  if (pokracovat())
  {
    printh(
      "- Existuje několik desítek předdefinovaných symbolů, které lze použít\n"
      "  pro řízení překladu nebo pro diagnostiku.\n"
      "- Jejich seznam lze vypsat pomocí parametrů překladače " KK("-dM -E") ".\n"
      "\n"
      KOD(
      "  echo | gcc -dM -E -\n"
      )
      "\n"
      "- Pro detekci operačních systémů není jednotný standard, ale nejvíce\n"
      "  podporované jsou tyto:\n"
      KOD(
      "  __unix__, __linux__\n"
      "  __FreeBSD__, __NetBSD__, __OpenBSD__\n"
      "  __APPLE__, __MACH__\n"
      "  __CYGWIN__, __CYGWIN32__, __MINGW32__, __MINGW64__\n"
      "  _WIN32, _WIN64\n"
      )
      "\n"
    );
  }

  /* Ještě existuje makro __COUNTER__, které postupně vrací čísla od 0.
  Při každém použití se hodnota zvětší o jedničku.
  Nenapadá mě rozumné využití, abych mohl napsat ukázku.
  Nejde to použít pro počítání iterací - je to makro, které funguje jen
  v době překladu.

  Kdybyste tím chtěli generovat unikátní identifikátory, nejde to dělat takto
  #define UNIKATNI_ID(nazev) nazev ## __COUNTER__

  UNIKATNI_ID(neco) ---> neco__COUNTER__

  To makro musí mít příležitost se rozvinout, takže se to musí dělat přes
  dokonce dvě pomocná makra...
  #define __SPOJ__(a,b) a ## b
  #define _SPOJ_(a,b) __SPOJ__(a,b)
  #define UNIKATNI_ID(nazev) _SPOJ_(nazev, __COUNTER__)

  UNIKATNI_ID(neco) ---> neco0
  UNIKATNI_ID(neco) ---> neco1
  UNIKATNI_ID(neco) ---> neco2
  */

  konecKapitoly(__func__);
}

////////////////////////////////////////////////////////////////////////////////


void direktivyPrekladace(void)
{
  zacatekKapitoly("Pragma direktivy překladače");

  printh(
    "- Preprocesor může pomocí příkazu " KK("#pragma") " řídit chod překladače.\n"
    "- Funguje to ale jen se specifickým překladačem.\n"
    "- Před použitím je potřeba si přečíst manuál daného překladače.\n"
    "\n"
    "- Přenositelný kód nemůže na těchto direktivách záviset!\n"
    "\n"
    "- Zde neuvádím plný seznam direktiv, jen jejich výběr. Zbytek viz dokumentace.\n"
    "\n"
  );

  if (navratNeboSekce("Poznámky, varování, chyby")) return;
  if (pokracovat())
  {
    printh(
      "\n"
      KOD(
      "  #pragma message \"TODO: Tohle je provizorní řešení. Oprav to!\"\n"
      "  #pragma GCC warning \"Tohle jen na vlastní nebezpečí!\"\n"
      "  #pragma GCC error \"Tohle přeložit nejde!\"\n"
      )
      "\n"
      "- Zprávy se objeví jako hlášení překladače v době překladu.\n"
      "- Příkaz " KK("#error") " je přímo ve standardu ISO C, takže by měl\n"
      "  být srozumitelný i jiným překladačům.\n"
      "\n"
    );
  }

  if (navratNeboSekce("Diagnostické direktivy")) return;
  if (pokracovat())
  {
    printh(
      "- Umožňují při překladu přepínat parametry překladače " KK("-W...") ", které\n"
      "  ovládají různé typy varování.\n"
      "\n"
      KOD(
      "#pragma GCC diagnostic warning \"-Wreturn-type\"\n"
      "#pragma GCC diagnostic error \"-Wreturn-type\"\n"
      "#pragma GCC diagnostic ignored \"-Wreturn-type\"\n"
      )
      "\n"
      "- Řídí chování daného přepínače bez ohledu na to, zda a jak byl použit\n"
      "  při volání překladače.\n"
      "- Přepínač " KK("warning") " zapíná dané varování.\n"
      "- Přepínač " KK("error") " udělá z varování chybové hlášení.\n"
      "- Přepínač " KK("ignored") " vypíná dané varování.\n"
      "\n"
      KOD(
      "#pragma GCC diagnostic push\n"
      "#pragma GCC diagnostic pop\n"
      )
      "\n"
      "- Přepínač " KK("push") " uloží aktuální nastavení všech přepínačů " KK("-W...") ".\n"
      "- Přepínač " KK("pop") " obnoví uložené nastavení přepínačů.\n"
      "\n"
    );
  }

  /* Vektorizace cyklů.
   - Pokud víte, že všechny iterace cyklu mohou proběhnout paraleleně, tedy
     nejsou na sobě nijak závislé, můžete cyklus vektorizovat.
   - Překlač pak cyklus provede pomocí SIMD instrukcí.

    void foo (int n, int *a, int *b, int *c)
    {
      int i, j;
    #pragma GCC ivdep
      for (i = 0; i < n; ++i)
        a[i] = b[i] + c[i];
    }

    - Výsledky ale silně závisí na verzi překladače a samotném kódu. Jednoduché
      cykly dokáže překladač optimalizovat sám a pak je rozdíl oproti tomuto
      zanedbatelný. V některých případech je to dokonce pomalejší.
  */


  if (navratNeboSekce("Použití v makrech")) return;
  if (pokracovat())
  {
    #define DO_PRAGMA(x) _Pragma (#x)
    #define TODO(x) DO_PRAGMA(message ("TODO - " #x))

    TODO(Nezapomeň koupit chleba a prodat bitcoiny!)

    printh(
      "- Pro použití v makrech existuje operátor " KK("_Pragma()") ".\n"
      "- Jeho parametrem je přesně to, co následuje za " KK("#pragma") ".\n"
      "\n"
      KOD(
      "#define DO_PRAGMA(x) _Pragma (#x)\n"
      "#define TODO(x) DO_PRAGMA(message (\"TODO - \" #x))\n"
      "\n"
      "TODO(Nezapomeň koupit chleba a prodat bitcoiny!)\n"
      )
      "\n"
    );
  }

  konecKapitoly(__func__);
}
