/* Encyklopedie jazyka C
 *
 * Modul s menu
 *
 * Tento zdrojový soubor modulu obsahuje implementaci pomocných podprogramů.
 *
 * 2023-08-20-11.09
 * 2023-07-31-22.52
 */

/* Copyright (c) David Martinek, 2023
 * david.martinek.78 (zavináč) gmail.com
 *
 * This work is licensed under a Creative Commons
 * Attribution-NonCommercial-ShareAlike 4.0 International License.
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 * Toto dílo je licencováno pole licence Creative Commons
 * Uveďte původ-Neužívejte dílo komerčně-Zachovejte licenci 4.0 Mezinárodní
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/deed.cs
 */

#include "menu.h"
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

/** Typ Tpolozka je položkou menu.
 * Tento typ je ukrytý.
 * Položky menu tvoří lineární seznam.
 */
typedef struct _polozka Tpolozka;

/** Specifikace položek v typu Tpolozka. */
struct _polozka
{
  char aktivator;        /**< Aktivační znak nabídky */
  // const je závazek, že se obsah nabídek nebude modifikovat, tudíž sem lze přiřadit konstantu
  const char * nabidka;  /**< Text nabídky */
  TFobsluha akce;        /**< Ukazatel na funkci */
  Tpolozka * dalsi;      /**< Ukazatel na další položku menu. */
};

/** Specifikace položek v typu Tmenu. Vnitřní položky tohoto typu jsou skryté. */
struct _menu
{
  const char *hlavicka; // závazek, že se obsah nabídek nebude modifikovat, tudíž sem lze přiřadit konstantu
  Tpolozka * prvni;
  Tpolozka * posledni;
  TFobsluha akcePredVolbou; // volá se po vypsání všech položek menu
};


Tmenu * menuVytvor(const char *hlavicka, TFobsluha akcePredVolbou)
{
  Tmenu * menu = malloc(sizeof(Tmenu));
  if (menu != NULL)
  {
    menu->hlavicka = hlavicka;
    menu->prvni = NULL;
    menu->posledni = NULL;
    menu->akcePredVolbou = akcePredVolbou;
  }
  return menu;
}

void menuZrus(Tmenu * menu)
{
  for (Tpolozka *p = menu->prvni; p != NULL; )
  {
    Tpolozka *a = p;
    p = p->dalsi;
    free(a);
  }
  free(menu);
}

/** Vytvoří novou položku menu.
 *
 * \param aktivator char Aktivační znak položky
 * \param nabidka const char* Text nabídky
 * \param akce TFobsluha Ukazatel na obslužnou funkci
 * \return Tpolozka* Vrací ukazatel na novou položku nebo NULL.
 */
Tpolozka * _novaPolozka(char aktivator, const char * nabidka, TFobsluha akce)
{
  Tpolozka * nova = malloc(sizeof(Tpolozka));
  if (nova != NULL)
  {
    nova->akce = akce;
    nova->aktivator = aktivator;
    nova->nabidka = nabidka;
    nova->dalsi = NULL;
  }
  return nova;
}

bool menuPridej(Tmenu * menu, char aktivator, const char * nabidka, TFobsluha akce)
{
  Tpolozka * nova = _novaPolozka(aktivator, nabidka, akce);
  if (nova == NULL) return false;

  if (menu->posledni == NULL)
  {
    menu->prvni = menu->posledni = nova;
    return true;
  }

  menu->posledni->dalsi = nova;
  menu->posledni = nova;
  return true;
}

/** Projde seznam položek menu a zobrací celé menu.
 * \param menu Tmenu* Ukazatel na menu
 */
void _menuZobrazit(Tmenu *menu)
{
  vycistiKonzoli();

  printg(menu->hlavicka);

  for (Tpolozka *p = menu->prvni; p != NULL; p = p->dalsi)
  {
    if (p->aktivator != '\0')
    {
      printg("  %c - ", p->aktivator);
    }
    printg(p->nabidka);
  }

  if (menu->akcePredVolbou != NULL)
  {
    menu->akcePredVolbou();
  }

  printg(
    "\n"
    "Q nebo Enter - Konec\n"
    "\n"
    VOLBA("Zadej písmeno volby: ")
  );

}

int menuVolba(Tmenu *menu)
{
  _menuZobrazit(menu);
  int volba = toupper(prvniPismenoNaRadku());
  vycistiKonzoli();

  if (volba == EOF || volba == '\n' || volba == 'Q')
  {
    return 'Q';
  }

  for (Tpolozka *p = menu->prvni; p != NULL; p = p->dalsi)
  {
    if (p->aktivator == volba)
    {
      p->akce();
      return volba;
    }
  }

  printf("Neznámá volba. Zkus to znovu.\n");
  return volba;
}






void malaRychlost(void)    { nastavVypraveciMod(true); prepniRychlostVypraveni(POMALA);  }
void stredniRychlost(void) { nastavVypraveciMod(true); prepniRychlostVypraveni(STREDNI); }
void velkaRychlost(void)   { nastavVypraveciMod(true); prepniRychlostVypraveni(RYCHLA);  }

void nastavRychlostVypraveni(void)
{
  Tmenu * menu = menuVytvor("Nastavení rychlosti vyprávění\n\n", zobrazMod);
  if (menu == NULL) return;

  menuPridej(menu, '0', "Přepni vyprávěcí mód\n", prepniVypraveciMod);
  menuPridej(menu, '1', "Velká rychlost\n", velkaRychlost);
  menuPridej(menu, '2', "Střední rychlost\n", stredniRychlost);
  menuPridej(menu, '3', "Malá rychlost \n", malaRychlost);

  do {
    printh(
      "Demonstrace nastavené rychlosti vyprávění:\n"
      "  Znám benzínovou studánku,\n"
      "  kde nejhlubší je les,\n"
      "  tam stačí škrtnout zápalku\n"
      "  a letíš do nebes.\n"
      "\n"
      "Pro pokračování stiskni Enter.\n"
    );
    pozerKonecRadku();
  } while (menuVolba(menu) != 'Q');

  menuZrus(menu);
}
