/* Encyklopedie jazyka C
 *
 * Modul s pomocnými podprogramy a makry.
 *
 * Tento zdrojový soubor modulu obsahuje implementaci pomocných podprogramů.
 *
 * 2023-08-20-11.09
 * 2023-07-31-22.52
 */

/* Copyright (c) David Martinek, 2021--2023
 * david.martinek.78 (zavináč) gmail.com
 *
 * This work is licensed under a Creative Commons
 * Attribution-NonCommercial-ShareAlike 4.0 International License.
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/
 *
 * Toto dílo je licencováno pole licence Creative Commons
 * Uveďte původ-Neužívejte dílo komerčně-Zachovejte licenci 4.0 Mezinárodní
 * (CC BY-NC-SA 4.0)
 * https://creativecommons.org/licenses/by-nc-sa/4.0/deed.cs
 */

#include "pomocny-modul.h"
#include "menu.h"
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdarg.h>

void vycistiKonzoli(void)
{
  int e = -1;
#if defined __unix__
  e = system("clear");
#elif defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
  e = system("cls");
#endif

  if (e != 0)
    printf("\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n\n");
}

void pozerKonecRadku(void)
{
  int c;
  // přeskočí vše do konce řádku
  while ( (c = getchar()) != '\n' && c != EOF)
  {} // prázdné tělo cyklu
}

int prvniPismenoNaRadku()
{
  int c;
  // přeskočí mezery
  while ( isblank(c = getchar()) )
  {} // prázdné tělo cyklu

  if (c != '\n')
    pozerKonecRadku(); // přečte a zahodí vše ostatní na řádku

  return c;

}


// Proměnná modulu - mimo modul není viditelná.
bool __preskocit = false;

bool pokracovat(void)
{
  if (__preskocit)
  {
    __preskocit = false;
    return false;
  }
  return true;
}

bool navratNeboSekce(char * nasledujiciSekce)
{
  while (1)
  {
    char sekce[2*81]; // České znaky v UTF-8 zabírají 2 bajty...
    char str[2*2*81];
    sprintg(sekce, "\n" SEKCE("%s"), nasledujiciSekce);
    sprintg(str,
      VOLBA(
      "### Následuje: %s\n"
      "### Zpět:Q, Přeskoč:P, Rychlost:0, Pokračuj:Enter "
      )
      , nasledujiciSekce
    );
    printg(str);
    int volba = tolower(prvniPismenoNaRadku());
    if (volba == '0')
    {
      vycistiKonzoli();
      nastavRychlostVypraveni();
      continue;
    }
    if (volba == 'p')
    {
      __preskocit = true;
      return false;
    }
    if (volba != 'q')
    {
      printh(sekce);
      return false;
    }
    break;
  }
  return true;
}


void konecKapitoly(const char * jmenoFunkce)
{
  if (jmenoFunkce != NULL)
    printh(
      "\n"
      "Pro další informace se podívej do zdrojového textu funkce " KK("%s") ".\n"
      , jmenoFunkce
    );

  printg(
    "\n"
    VOLBA("### Pro návrat do menu stiskni Enter.") "\n"
  );
  pozerKonecRadku(); // čeká, dokud uživatel nestiskne Enter.
}



// Tohle není obyčejný hlavičkový soubor. Je překódovaný do CP852. Neupravovat!
// Je v něm definice konstantní proměnné. Nesmí se proto vkládat do více modulů!
#include "utf8nacp852.h"

#define UTF8_PREFIX_MIN (-61)
#define UTF8_PREFIX_MAX (-59)

void _prevedUTF8NaCP852(char *vyslednyString, const char *utfString)
{
  int j = 0;
  for (int i = 0; utfString[i] != 0; ++i, ++j)
  {
    int z1 = utfString[i];
    if (z1 >= UTF8_PREFIX_MIN && z1 <= UTF8_PREFIX_MAX) // pozor magie! nesahat!
    {
      // V UTF-8 jsou české znaky dvoubajtové. Tímto se přepočítají na indexy
      // v tabulce UTF8_NA_CP852.
      z1 -= UTF8_PREFIX_MIN; // 61 - (-61) => 0
      int z2 = utfString[++i]+128;
      vyslednyString[j] = UTF8_NA_CP852[z1][z2];
    }
    else
      vyslednyString[j] = z1;
  }
  vyslednyString[j] = '\0';
}

void zacatekKapitoly(const char *nazev)
{
  char radek1[101];
  char radek2[101];

#ifdef __unix__
  sprintf(radek2, "     %s     ", nazev);
#else
  char czNazev[2*strlen(nazev)+1];
  _prevedUTF8NaCP852(czNazev, nazev);
  sprintf(radek2, "     %s     ", czNazev);
#endif // __unix__

  int n = strlen(radek2);

  for (int i = 0; i < n; ++i) { radek1[i] = '*'; }
  radek1[n] = '\0';

  printh( NADPIS("%s\n%s\n%s\n") "\n" , radek1, radek2, radek1);
}



/**
 * Funguje jako printf, ale převede formátovací řetězec do CP852.
 * Používá tabulku UTF8_NA_CP852 ze souboru "utf8nacp852.h".
 *
 * Varianta s parametrem typu va_list. Určeno pro volání z funkcí s proměnným
 * počtem parametrů, tj. seznamem parametrů končícím třemi tečkami ...
 */
int _vprintg(const char * format, va_list args)
{
  char upravenyFormat[strlen(format)+1];
  _prevedUTF8NaCP852(upravenyFormat, format);
  int znaku = vprintf(upravenyFormat, args);

  return znaku;
}


/**
 * Funguje jako printf, ale převede formátovací řetězec do CP852.
 * Používá tabulku UTF8_NA_CP852 ze souboru "utf8nacp852.h".
 *
 * Funkce s proměnným počtem parametrů - určena pro uživatelské použití.
 */
int printg(const char * format, ...)
{
  va_list args;
  va_start(args, format);

  int znaku = 0;

// Poznámka: předpokládám, že na unixu je nastavena lokalizace cs_CZ.UTF-8
// a ve Windows konzole používá CP852.
#ifdef __unix__
  // v Unixu/Linuxu
  znaku = vprintf(format, args);
#else
  // ve Windows
  znaku = _vprintg(format, args);
#endif // __unix__

  va_end(args);
  return znaku;
}


/**
 * Funguje jako vsprintf, ale překóduje formátovací řetězec z UTF-8 do CP852.
 *
 * Varianta s parametrem typu va_list. Určeno pro volání z funkcí s proměnným
 * počtem parametrů, tj. seznamem parametrů končícím třemi tečkami ...
 */
int _vsprintg(char *str, const char * format, va_list args)
{
  // Text v CP852 bude vždy zabírat méně bajtů než text v UTF-8, takže by
  // tato velikost měla vždy stačit.
  char upravenyFormat[strlen(format)+1];
  _prevedUTF8NaCP852(upravenyFormat, format);
  int znaku = vsprintf(str, upravenyFormat, args);

  return znaku;
}

/**
 * Funguje jako sprintf, ale převede formátovací řetězec do CP852.
 * Používá tabulku UTF8_NA_CP852 ze souboru "utf8nacp852.h".
 *
 * Funkce s proměnným počtem parametrů - určena pro uživatelské použití.
 */
int sprintg(char * str, const char * format, ...)
{
  va_list args;
  va_start(args, format);

  int znaku = 0;

// Poznámka: předpokládám, že na unixu je nastavena lokalizace cs_CZ.UTF-8
// a ve Windows konzole používá CP852.
#ifdef __unix__
  // v Unixu/Linuxu
  znaku = vsprintf(str, format, args);
#else
  // ve Windows
  znaku = _vsprintg(str, format, args);
#endif // __unix__

  va_end(args);
  return znaku;
}


#if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
  #include <windows.h>
#else
  // kvůli usleep - ale ta už je zastaralá => používám nanosleep
  //#include <unistd.h>

  // kvůli nanosleep v unixových systémech
  #define __USE_POSIX199309 1
  #include <time.h>
#endif


void povolBarevnyVystup(void)
{
  // Protože Windows musí mít vždycky nějaké extra buřty...
  // V Linuxu bude tohle prázdná funkce. Tam terminály fungují standardním
  // způsobem.
  #if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
    HANDLE hOut = GetStdHandle(STD_OUTPUT_HANDLE);
    DWORD dwMode = 0;
    GetConsoleMode(hOut, &dwMode);
//    dwMode |= ENABLE_VIRTUAL_TERMINAL_PROCESSING;
    dwMode |= 0x0004;
    SetConsoleMode(hOut, dwMode);

    // References:
    //SetConsoleMode() and ENABLE_VIRTUAL_TERMINAL_PROCESSING?
    //https://stackoverflow.com/questions/38772468/setconsolemode-and-enable-virtual-terminal-processing

    // Windows console with ANSI colors handling
    // https://superuser.com/questions/413073/windows-console-with-ansi-colors-handling
#endif
}




// Proměnná modulu - mimo modul není viditelná.
bool _modVypraveni = true;


void nastavVypraveciMod(bool vypravet)
{
  _modVypraveni = vypravet;
}

void prepniVypraveciMod(void)
{
  _modVypraveni = !_modVypraveni;
}

bool jeVypraveciMod(void)
{
  return _modVypraveni;
}


int _rychlost = STREDNI;

void prepniRychlostVypraveni(int rychlost)
{
  _rychlost = rychlost;
}



void zobrazMod(void)
{
  printg("\n");
  if (_modVypraveni)
  {
    printg("Mód: VYPRÁVĚCÍ/statický\n");
    printg("Rychlost vyprávění: ");
    switch (_rychlost)
    {
      case POMALA:  printg("POMALÁ\n");   break;
      case STREDNI: printg("STŘEDNÍ\n");  break;
      case RYCHLA:  printg("RYCHLÁ\n");   break;
      default:      printg("ABSURDNÍ\n"); break;
    }
  }
  else
  {
    printg("Mód: vyprávěcí/STATICKÝ\n");
  }
}



/** Čeká zadaný počet milisekund.
 *
 * V unixu a ve Windows se dělá čekání jinak. V unixu používá funkci nanosleep,
 * ve Windows funkci Sleep.
 *
 * \param milisekundy unsigned int Počet milisekund, které je potřeba čekat.
 * \return void
 */
void _sleepcp(unsigned int milisekundy)
{
//    #if defined WIN32 || defined _WIN32 || defined WIN64 || defined _WIN64
  #ifdef __unix__
//    usleep(milliseconds * 1000);  // usleep je zastaralá
    struct timespec ts;
    ts.tv_sec = milisekundy / 1000; // milisekundy na sekundy
    ts.tv_nsec = (milisekundy % 1000) * 1000000; // Zbývající milisekundy na nanosekundy
    nanosleep(&ts, NULL);
  #else
    Sleep(milisekundy);
  #endif // __unix__
}


/** Tiskne text ve vyprávěcím módu.
 *
 * Po každém vytištěném znaku čeká znakms milisekund, po každé mezeře mezerams
 * milisekund a na konci každého řádku radekms milisekund.
 *
 * Tento mód pomáhá čtenáři udržovat pozornost.
 *
 * Poznámka 1: Pokud je text v parametru str v UTF-8, linuxový terminál umí
 * tisknout vícebajtové znaky vytištěné dvěma voláními putchar jako jeden znak.
 * Dělá to tak, že interpretuje vnitřní buffer. Pokud ale dojde k nucenému
 * vyprázdnění bufferu uprostřed takového vícebajtového (v češtině 2 bajty)
 * znaku pomocí fflush, může to na některých systémech vést k nepředvídatelnému
 * chování. - V Ubuntu to funguje bez problémů, ale když jsem to zkoušel na
 * replit.com, došlo k zaseknutí výstupního streamu. Řeším to tak, že detekuji
 * první bajt českého znaku a v takovém případě volání fflush nedovolím.
 * Tohle řešení funguje spolehlivě i na replit.com.
 *
 * Poznámka 2: Při tisku bez fflush se řádek vytiskne až po vytištění konce
 * řádku, protože je výstup na stdout bufferovaný.
 * Volání fflush(stdout) ale nemusí fungovat všude, protože fflush má podle
 * dokumentace fungovat jen na streamech v user-space. V unixu (Ubuntu, replit)
 * to funguje, ve Windows ne. Terminál Windows tiskne výstup po jednotlivých
 * znacích sám od sebe - asi nepoužívá řádkový buffer...
 *
 * Poznámka 3: Tato funkce je tak trochu vúdú. Robustnější řešení by asi
 * mohla poskytnout knihovna ncurses nebo něco podobného.
 *
 * \param str char*     Text, který má být odvyprávěn.
 * \param znakms int    Čekání po znaku
 * \param mezerams int  Čekání po slově
 * \param radekms int   Čekání po řádku
 * \return void
 */
void _printChat(char *str, int znakms, int mezerams, int radekms)
{
  int minulyznak = '\0';

  // Na replit.com nevypadá dobře, když se dělá fflush po každém znaku.
  // Proměnná mindelka slouží k počítání délky posloupnosti znaků, po které
  // se fflush může provést.
  int mindelka = 0;

  for (int i = 0; str[i] != '\0'; ++i)
  {
    int z1 = str[i];
    putchar(str[i]);

    mindelka += 1;

    int milis = znakms;

    // K této hodnotě jsem dospěl experimentováním. Příliš malá hodnota
    // znamená příliš časté volání fflush a na replit.com to způsobí paradoxně
    // zpomalení. Příliš velká hodnota pak zase vypadá trhaně.
    if (mindelka >= 4)
    {
      if (!(z1 >= UTF8_PREFIX_MIN && z1 <= UTF8_PREFIX_MAX)) // pozor magie! nesahat!
      {
        // Ukazuje se, že některé systémy nemají rády, když se udělá fflush
        // uprostřed UTF-8 znaku. Například na replit.com pak kompletně
        // zatuhne výstup.
        fflush(stdout);
        mindelka = 0;
      }
      else if (!isblank(minulyznak) && isblank(str[i]))
      {
        milis = mezerams;
        fflush(stdout);
        mindelka = 0;
      }
    }

    if (minulyznak != '\n' && str[i] == '\n')
    {
      milis = radekms;
      mindelka = 0;
      fflush(stdout);
    }

    minulyznak = str[i];
    _sleepcp(milis);
  }
}


int printh(const char * format, ...)
{
  va_list args;
  va_start(args, format);

  int znaku;
  if (_modVypraveni)
  {
    char str[500*81] = {'\0'};
    #ifdef __unix__
      // v Unixu/Linuxu
      znaku = vsprintf(str, format, args);
    #else
      // ve Windows
      znaku = _vsprintg(str, format, args);
    #endif // __unix__
    _printChat(str, _rychlost, _rychlost*10, _rychlost*20);
  }
  else
  {
    #ifdef __unix__
      // v Unixu/Linuxu
      znaku = vprintf(format, args);
    #else
      // ve Windows
      znaku = _vprintg(format, args);
    #endif // __unix__
  }

  va_end(args);
  return znaku;
}
